/* See COPYRIGHT for copyright information. */

#include <inc/x86.h>
#include <inc/error.h>
#include <inc/string.h>
#include <inc/assert.h>

#include <kern/env.h>
#include <kern/pmap.h>
#include <kern/trap.h>
#include <kern/syscall.h>
#include <kern/console.h>
#include <kern/sched.h>
#include <kern/kernbin.h>


/**********************************
 * SYSTEM CALLS FOR LAB 3, PART 1 *
 *                                *
 **********************************/

// Print a string to the system console.
// The system call returns 0.
static void
sys_cputs(const char* s)
{
	printf("%s", s);
}

// Read a character from the system console.
// Returns the character.
static int
sys_cgetc(void)
{
	int c;

	// The cons_getc() primitive doesn't wait for a character,
	// but the sys_cgetc() system call does.
	while ((c = cons_getc()) == 0)
		; /* spin */

	return c;
}

// Return the current environment's envid.
static envid_t
sys_getenvid(void)
{
	return curenv->env_id;
}

// Destroy a given environment (possibly the currently running environment).
//
// Returns 0 on success, < 0 on error.  Errors are:
//	-E_BAD_ENV if envid is not a running environment,
//		or the caller doesn't have permission to change envid.
static int
sys_env_destroy(envid_t envid)
{
	int r;
	struct Env *e;

	if ((r = envid2env(envid, &e, 1)) < 0)
		return r;
	if (e == curenv)
		printf("[%08x] exiting gracefully\n", curenv->env_id);
	else
		printf("[%08x] destroying %08x\n", curenv->env_id, e->env_id);
	env_destroy(e);
	return 0;
}



/********************************************************
 * SYSTEM CALLS FOR LAB 3, PART 3                       *
 *                                                      *
 * No need to implement these until you get to Part 3.  *
 *                                                      *
 ********************************************************/

// Exercise 8:
// Deschedule current environment and pick a different one to run.
// The system call returns 0.
static void
sys_yield(void)
{
	sched_yield();
}

// Exercise 9:
// Allocate a new environment.
// Returns envid of new environment, or < 0 on error.  Errors are:
//	-E_NO_FREE_ENV if no free environment is available.
static envid_t
sys_exofork(void)
{
	// Create the new environment with env_alloc(), from kern/env.c.
	// It should be left as env_alloc created it, except that
	// status is set to ENV_NOT_RUNNABLE, and the register set is copied
	// from the current environment -- but tweaked so sys_exofork
	// will appear to return 0.
	//
	// Hint: Your code in env_run() shows how to copy a register set.
	
	// LAB 3 PART 3: Your code here.

	panic("sys_exofork not implemented");
}

// Exercise 9:
// Set envid's env_status to status, which must be ENV_RUNNABLE
// or ENV_NOT_RUNNABLE.
//
// Returns 0 on success, < 0 on error.  Errors are:
//	-E_BAD_ENV if envid is not a running environment,
//		or the caller doesn't have permission to change envid.
//	-E_INVAL if status is not a valid status for an environment.
static int
sys_env_set_status(envid_t envid, int status)
{
  	// Hint: Use the 'envid2env' function from kern/env.c to translate an
  	// envid to a struct Env.
	// You should set envid2env's third argument to 1, which will
	// check whether the current environment has permission to set
	// envid's status.
	
	// LAB 3 PART 3: Your code here.
	panic("sys_env_set_status not implemented");
}

// Exercise 9:
// Allocate a page of memory and map it at 'va' with permission
// 'perm' in the address space of 'envid'.
// The page's contents are set to 0.
// If a page is already mapped at 'va', that page is unmapped as a
// side effect.
//
// perm -- PTE_U | PTE_P must be set, PTE_AVAIL | PTE_W may or may not be set,
//         but no other bits may be set.
//
// Return 0 on success, < 0 on error.  Errors are:
//	-E_BAD_ENV if envid is not a running environment,
//		or the caller doesn't have permission to change envid.
//	-E_INVAL if va >= UTOP, or va is not page-aligned.
//	-E_INVAL if perm is inappropriate (see above).
//	-E_NO_MEM if there's no memory to allocate the new page,
//		or to allocate any necessary page tables.
static int
sys_page_alloc(envid_t envid, uintptr_t va, int perm)
{
	// Hint: This function is a wrapper around page_alloc() and
	//   page_insert() from kern/pmap.c.
	//   Most of the new code you write should be to check the
	//   parameters for correctness.
	//   If page_insert() fails, remember to free the page you
	//   allocated!

	// LAB 3 PART 3: Your code here.
	panic("sys_page_alloc not implemented");
}

// Exercise 9:
// Map the page of memory at 'srcva' in srcenvid's address space
// at 'dstva' in dstenvid's address space with permission 'perm'.
// Perm has the same restrictions as in sys_page_alloc, except
// that it also must not grant write access to a read-only
// page.
//
// Return 0 on success, < 0 on error.  Errors are:
//	-E_BAD_ENV if srcenvid and/or dstenvid is not a running environment,
//		or the caller doesn't have permission to change one of them.
//	-E_INVAL if srcva >= UTOP or srcva is not page-aligned,
//		or dstva >= UTOP or dstva is not page-aligned.
//	-E_INVAL is srcva is not mapped in srcenvid's address space.
//	-E_INVAL if perm is inappropriate (see sys_page_alloc).
//	-E_INVAL if (perm & PTE_W), but srcva is read-only in srcenvid's
//		address space.
//	-E_NO_MEM if there's no memory to allocate the new page,
//		or to allocate any necessary page tables.
static int
sys_page_map(envid_t srcenvid, uintptr_t srcva,
	     envid_t dstenvid, uintptr_t dstva,
	     int perm)
{
	// Hint: This function is a wrapper around page_lookup() and
	//   page_insert() from kern/pmap.c.
	//   Again, most of the new code you write should be to check the
	//   parameters for correctness.
	//   Use the third argument to page_lookup() to
	//   check the current permissions on the page.

	// LAB 3 PART 3: Your code here.
	panic("sys_page_map not implemented");
}

// Exercise 9:
// Unmap the page of memory at 'va' in the address space of 'envid'.
// If no page is mapped, the function silently succeeds.
//
// Return 0 on success, < 0 on error.  Errors are:
//	-E_BAD_ENV if envid is not a running environment,
//		or the caller doesn't have permission to change envid.
//	-E_INVAL if va >= UTOP, or va is not page-aligned.
static int
sys_page_unmap(envid_t envid, uintptr_t va)
{
	// Hint: This function is a wrapper around page_remove().
	
	// LAB 3 PART 3: Your code here.
	panic("sys_page_unmap not implemented");
}


// Lab 4 Exercise 1:
// Set the page fault upcall for 'envid' by modifying the corresponding struct
// Env's 'env_pgfault_upcall' field.  When 'envid' causes a page fault, the
// kernel will push a fault record onto the exception stack, then branch to
// 'func'.
//
// Returns 0 on success, < 0 on error.  Errors are:
//	-E_BAD_ENV if envid is not a running environment,
//		or the caller doesn't have permission to change envid.
static int
sys_set_pgfault_upcall(envid_t envid, uintptr_t upcall)
{
	// LAB 4: Your code here.
	panic("sys_set_pgfault_upcall not implemented");
}

// Lab 4 Exercise 8:
// Block until a value is ready.  Record that you want to receive
// using the env_ipc_recving and env_ipc_dstva fields of struct Env,
// mark yourself not runnable, and then give up the CPU.
//
// If 'dstva' is < UTOP, then you are willing to receive a page of data.
// 'dstva' is the virtual address at which the sent page should be mapped.
//
// This function only returns on error, but the system call will eventually
// return 0 on success.
// Return < 0 on error.  Errors are:
//	-E_INVAL if dstva < UTOP but dstva is not page-aligned.
static int
sys_ipc_recv(uintptr_t dstva)
{
	// LAB 4: Your code here.
	panic("sys_ipc_recv not implemented");
	return 0;
}

// Lab 4 Exercise 8:
// Try to send 'value' to the target env 'envid'.
// If va != 0, then also send page currently mapped at 'va',
// so that receiver gets a duplicate mapping of the same page.
//
// The send fails with a return value of -E_IPC_NOT_RECV if the
// target has not requested IPC with sys_ipc_recv.
//
// Otherwise, the send succeeds, and the target's ipc fields are
// updated as follows:
//    env_ipc_recving is set to 0 to block future sends;
//    env_ipc_from is set to the sending envid;
//    env_ipc_value is set to the 'value' parameter;
//    env_ipc_perm is set to 'perm' if a page was transferred, 0 otherwise.
// The target environment is marked runnable again.
//
// If the sender sends a page but the receiver isn't asking for one,
// then no page mapping is transferred, but no error occurs.
//
// Returns 0 on success where no page mapping occurs,
// 1 on success where a page mapping occurs, and < 0 on error.
// Errors are:
//	-E_BAD_ENV if envid is not a running environment.
//		(No need to check permissions.)
//	-E_IPC_NOT_RECV if envid is not currently blocked in sys_ipc_recv,
//		or another environment managed to send first.
//	-E_INVAL if srcva < UTOP but srcva is not page-aligned.
//	-E_INVAL if srcva < UTOP and perm is inappropriate
//		(see sys_page_alloc).
//	-E_INVAL if srcva < UTOP but srcva is not mapped in the caller's
//		address space.
//	-E_NO_MEM if there's not enough memory to map srcva in envid's
//		address space.
static int
sys_ipc_try_send(envid_t envid, uint32_t value, uintptr_t srcva, unsigned perm)
{
	// LAB 4: Your code here.
	panic("sys_ipc_try_send not implemented");
}

// Allocate a page of memory and map it at 'va' with permission
// 'perm' (which must be read-only) in the address space of 'envid'.
// The page's contents are set to data from the kernel binary named 'name'
// starting at binary offset 'offset'.
// Thus, multiple calls to sys_kernbin_page_alloc will map the entire
// named ELF binary.
// Any bytes past the length of the kernel binary are set to 0.
// If a page is already mapped at 'va', that page is unmapped as a
// side effect.
// Multiple calls to sys_kernbin_page_alloc() for the same binary/offset
// pair can return mappings to the same page (i.e. there is a cache).
//
// Returns the size of the kernel binary on success, < 0 on error.  Errors are:
//	-E_BAD_ENV if envid is not a running environment,
//		or the caller doesn't have permission to change envid.
//	-E_INVAL if va >= UTOP, or va is not page-aligned.
//	-E_INVAL if perm is inappropriate (see sys_page_alloc).
//	-E_INVAL if perm contains PTE_W.
//	-E_INVAL if name does not name a valid kernel binary.
//	-E_INVAL if offset is not page-aligned.
//	-E_NO_MEM if there's no memory to allocate the new page,
//		or to allocate any necessary page tables.
static ssize_t
sys_kernbin_page_alloc(envid_t envid, const char* name, uint32_t offset, uintptr_t va, unsigned perm)
{
	struct Env* env;
	struct Kernbin* kernbin;
	struct Page* pg;
	int retval;
	int pgoff;
	size_t size;
	
	if ((retval = envid2env(envid, &env, 1)) < 0)
		return retval;

	// Find the named kernbin structure.  Protect the binary name!
	name = TRUP(name);
	page_fault_mode = PFM_KILL;
	for (kernbin = kernbins; kernbin->name; kernbin++)
		if (strcmp(kernbin->name, name) == 0)
			break;
	page_fault_mode = PFM_NONE;
	if (!kernbin->name)
		return -E_INVAL;

	// Check the offset and destination VA
	if ((offset & (PGSIZE - 1))
	    || (va & (PGSIZE - 1))
	    || va >= UTOP
	    || (~perm & (PTE_U | PTE_P))
	    || (perm & ~(PTE_U | PTE_P | PTE_AVAIL)))
		return -E_INVAL;

	// Check the page cache
	pgoff = offset >> PGSHIFT;
	if (pgoff < KERNBIN_MAXPAGES && (pg = kernbin->pages[pgoff])) {
		if ((retval = page_insert(env->env_pgdir, pg, va, perm)) < 0)
			return retval;
		return kernbin->size;
	}
	
	// Allocate a page and map it at va
	if ((retval = page_alloc(&pg)) < 0)
		return retval;
	if ((retval = page_insert(env->env_pgdir, pg, va, perm)) < 0) {
		page_free(pg);
		return retval;
	}
	
	// Copy the relevant portion of the binary
	if (offset < kernbin->size) {
		size = PGSIZE;
		if (offset + size > kernbin->size)
			size = kernbin->size - offset;
		memcpy((void*) page2kva(pg), kernbin->data + offset, size);
	} else
		size = 0;

	// Rease remaining portion of page
	memset((uint8_t*) page2kva(pg) + size, 0, PGSIZE - size);

	// Cache the page (and don't lose the reference)
	if (pgoff < KERNBIN_MAXPAGES) {
		pg->pp_ref++;
		kernbin->pages[pgoff] = pg;
	}
	
	// Return the binary size
	return kernbin->size;
}


// Set envid's trap frame to 'tf'.
// tf is modified to make sure that user environments always run at code
// protection level 3 (CPL 3), and with interrupts enabled.
//
// Hint: You must do something special if 'envid' is the current environment!
//
// Returns 0 on success, < 0 on error.  Errors are:
//	-E_BAD_ENV if envid is not a running environment,
//		or the caller doesn't have permission to change envid.
static int
sys_set_trapframe(envid_t envid, struct Trapframe* tf)
{
	int retval;
	struct Env* e;
	struct Trapframe ltf;

	page_fault_mode = PFM_KILL;
	ltf = *TRUP(tf);
	page_fault_mode = PFM_NONE;

	ltf.tf_eflags |= FL_IF;
	ltf.tf_cs |= 3;

	if ((retval = envid2env(envid, &e, 1)) < 0)
		return retval;
	if (e == curenv)
		*UTF = ltf;
	else
		e->env_tf = ltf;
	return 0;
}


// Dispatches to the correct kernel function, passing the arguments.
register_t
syscall(register_t sn, register_t a1, register_t a2, register_t a3, register_t a4, register_t a5)
{
	// printf("syscall %d %x %x %x from env %08x\n", sn, a1, a2, a3, curenv->env_id);

	// LAB 3: Your code here.

	panic("syscall not implemented");
}

